/*
 * This file is part of the vpnd2 project
 *
 * based on http://burtleburtle.net/bob/c/lookup3.c
 * lookup3.c, by Bob Jenkins, May 2006, Public Domain.
 */

#ifndef HASH_H_INCLUDED
#define HASH_H_INCLUDED

#include <stdint.h>

#define hash_rot(x,k) (((x)<<(k)) | ((x)>>(32-(k))))

#define hash_mix(a,b,c) \
{ \
  a -= c;  a ^= hash_rot(c, 4);  c += b; \
  b -= a;  b ^= hash_rot(a, 6);  a += c; \
  c -= b;  c ^= hash_rot(b, 8);  b += a; \
  a -= c;  a ^= hash_rot(c,16);  c += b; \
  b -= a;  b ^= hash_rot(a,19);  a += c; \
  c -= b;  c ^= hash_rot(b, 4);  b += a; \
}

#define hash_final(a,b,c) \
{ \
  c ^= b; c -= hash_rot(b,14); \
  a ^= c; a -= hash_rot(c,11); \
  b ^= a; b -= hash_rot(a,25); \
  c ^= b; c -= hash_rot(b,16); \
  a ^= c; a -= hash_rot(c,4);  \
  b ^= a; b -= hash_rot(a,14); \
  c ^= b; c -= hash_rot(b,24); \
}

static inline uint32_t hash_ipv4(uint32_t saddr,uint32_t daddr,uint16_t sport,
	uint16_t dport)
{
	union
	{
		uint16_t data[2];
		uint32_t ports;
	} p;

	p.data[0]=sport;
	p.data[1]=dport;
	p.ports+=0xdeadbefb;
	saddr+=0xdeadbefb;
	daddr+=0xdeadbefb;
	hash_final(saddr,daddr,p.ports);
	return p.ports;
}

static inline uint32_t __hash_ipv6(uint32_t *saddr,uint32_t *daddr,
	uint16_t sport,uint16_t dport)
{
	uint32_t a,b,c;
	union
	{
		uint16_t data[2];
		uint32_t ports;
	} p;

	p.data[0]=sport;
	p.data[1]=dport;
	a=b=c=0xdeadbf13;
	a+=saddr[0];
	b+=saddr[1];
	c+=saddr[2];
	hash_mix(a,b,c);
	a+=saddr[3];
	b+=daddr[0];
	c+=daddr[1];
	hash_mix(a,b,c);
	a+=daddr[2];
	b+=daddr[3];
	c+=p.ports;
	hash_final(a,b,c);
	return c;
}

#define hash_ipv6(a,b,c,d) __hash_ipv6((uint32_t *)(a),(uint32_t *)(b),(c),(d))

#endif
