/*
 * This file is part of the vpnd2 project
 *
 * (C) 2017 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2 or, at
 * your choice, any later version of this license.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <poll.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include "misc.h"
#include "sctp_om.h"
#include "helper.h"

int helperconnect(void *helper,void *conf,void *envhost,int envport,int timeout,
	int packet,int efd)
{
	int pp[2];
	struct pollfd p[2];
	unsigned char bfr[32];

	if(socketpair(AF_UNIX,
		(packet?SOCK_SEQPACKET:SOCK_STREAM)|SOCK_NONBLOCK|SOCK_CLOEXEC,
		0,pp))goto err1;
	sctpomlock();
	switch(fork())
	{
	case -1:sctpomunlock();
		goto err2;
	case 0:	close(pp[0]);
		if(pp[1]<2)exit(1);
		if(dup2(pp[1],0)==-1)exit(1);
		if(fcntl(0,F_SETFL,O_RDWR))exit(1);
		if(dup2(pp[1],1)==-1)exit(1);
		if(fcntl(1,F_SETFL,O_RDWR))exit(1);
		close(pp[1]);
		if(envport)
		{
			sprintf((char *)bfr,"%d",envport);
			if(setenv("VPN_PORT",(char *)bfr,1))exit(1);
		}
		else unsetenv("VPN_PORT");
		if(!envhost)unsetenv("VPN_HOST");
		else if(setenv("VPN_HOST",envhost,1))exit(1);
		munlockall();
		execl(helper,helper,conf,NULL);
		exit(1);
	default:sctpomunlock();
	}

	p[0].fd=efd;
	p[0].events=POLLIN;
	p[1].fd=pp[0];
	p[1].events=POLLIN;
	if(poll(p,2,timeout*1000)<1)goto err2;
	if(p[0].revents&POLLIN)goto err2;
	if(!(p[1].revents&POLLIN))goto err2;
	if(read(pp[0],bfr,3)!=3)goto err2;
	if(strncmp((char *)bfr,"OK\n",3))goto err2;

	close(pp[1]);
	return pp[0];

err2:	close(pp[0]);
	close(pp[1]);
err1:	return -1;
}

void helperclose(int fd)
{
	close(fd);
}

HOT int helpersend(int fd,void *data,int len)
{
	int l;

	if(UNLIKELY((l=write(fd,data,len))==-1))switch(errno)
	{
	case EINTR:
	case EAGAIN:
		return 0;
	default:return -1;
	}
	return l;
}

HOT int helperrecv(int fd,void *data,int max)
{
	int l;

	if(UNLIKELY((l=read(fd,data,max))==-1))switch(errno)
	{
	case EINTR:
	case EAGAIN:
		return 0;
	default:return -1;
	}
	return l;
}

HOT int helperpktsend(int fd,void *data,int len)
{
	int l;

	if(UNLIKELY((l=send(fd,data,len,MSG_EOR))==-1))switch(errno)
	{
	case EINTR:
	case EAGAIN:
		return 0;
	default:return -1;
	}
	if(UNLIKELY(l!=len))return -1;
	return l;
}

HOT int helperpktrecv(int fd,void *data,int max)
{
	int l;

	if(UNLIKELY((l=recv(fd,data,max,MSG_NOSIGNAL|MSG_WAITALL))==-1))
		switch(errno)
	{
	case EINTR:
	case EAGAIN:
		return 0;
	default:return -1;
	}
	return l;
}
