/*
 * This file is part of the vpnd2 project
 *
 * (C) 2017 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2 or, at
 * your choice, any later version of this license.
 */

#include <sys/stat.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#ifdef DEBUG
#include <stdio.h>
#endif
#include "misc.h"
#include "crypto.h"
#include "keyio.h"

#ifdef DEBUG
#define ERR(a) do							\
{									\
	fprintf(stderr,"KeyIO abort: %s line %d\n",__FILE__,__LINE__);	\
	goto a;								\
} while(0)
#else
#define ERR(a)	goto a
#endif

COLD int genkey(void *h,char *fn,int curve,void *passwd,int plen)
{
	int r=-1;
	int fd;
	int len;
	void *key;
	void *bfr;
	void *pem;

	len=strlen(fn);
	if(!len||fn[len-1]=='/'||fn[len-1]=='.')ERR(err1);
	if(!(key=crypto_genkey(h,curve)))ERR(err1);
	if(!(bfr=malloc(strlen(fn)+5)))ERR(err2);
	strcpy(bfr,fn);
	strcat(bfr,".key");
	if(!(pem=crypto_make_pem_key(h,key,&len,passwd,plen)))ERR(err3);
	if((fd=open(bfr,O_WRONLY|O_TRUNC|O_CREAT|O_CLOEXEC,0600))==-1)ERR(err4);
	if(write(fd,pem,len)!=len)ERR(err5);
	close(fd);
	crypto_memclear(h,pem,len);
	free(pem);
	strcpy(bfr,fn);
	strcat(bfr,".pub");
	if(!(pem=crypto_make_pem_pub(h,key,&len)))ERR(err3);
	if((fd=open(bfr,O_WRONLY|O_TRUNC|O_CREAT|O_CLOEXEC,0644))==-1)ERR(err4);
	if(write(fd,pem,len)!=len)ERR(err5);
	r=0;
err5:	close(fd);
err4:	crypto_memclear(h,pem,len);
	free(pem);
err3:	free(bfr);
err2:	crypto_free_key(h,key);
err1:	return r;
}

COLD void *loadkey(void *h,void *fn,void *passwd,int plen)
{
	int fd;
	int len;
	struct stat stb;
	void *bfr;
	void *key=NULL;

	if(!stat(fn,&stb)&&!S_ISREG(stb.st_mode))ERR(err1);
	if(!(bfr=malloc(stb.st_size)))ERR(err1);
	if((fd=open(fn,O_RDONLY|O_CLOEXEC))==-1)ERR(err2);
	len=read(fd,bfr,stb.st_size);
	close(fd);
	if(len!=stb.st_size)ERR(err3);
	key=crypto_read_pem_key(h,bfr,len,passwd,plen);
err3:	crypto_memclear(h,bfr,stb.st_size);
err2:	free(bfr);
err1:	return key;
}

void *loadpub(void *h,void *fn,int *publen)
{
	int fd;
	int len;
	struct stat stb;
	void *bfr;
	void *pub=NULL;

	if(!stat(fn,&stb)&&!S_ISREG(stb.st_mode))ERR(err1);
	if(!(bfr=malloc(stb.st_size)))ERR(err1);
	if((fd=open(fn,O_RDONLY|O_CLOEXEC))==-1)ERR(err2);
	len=read(fd,bfr,stb.st_size);
	close(fd);
	if(len!=stb.st_size)ERR(err2);
	pub=crypto_read_pem_pub(h,bfr,len,publen);
err2:	free(bfr);
err1:	return pub;
}

int savepub(void *h,void *fn,void *pub,int publen)
{
	int fd;
	int plen;
	void *pem;

	if(!(pem=crypto_write_pem_pub(h,pub,publen,&plen)))ERR(err1);
	if((fd=open(fn,O_WRONLY|O_TRUNC|O_CREAT|O_CLOEXEC,0644))==-1)ERR(err2);
	if(write(fd,pem,plen)!=plen)goto err3;
	close(fd);
	return 0;

err3:	close(fd);
	unlink(fn);
err2:	free(pem);
err1:	return -1;
}
