/*
 * This file is part of the vpnd2 project
 *
 * (C) 2017 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2 or, at
 * your choice, any later version of this license.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <netinet/tcp.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>

static int v4test(struct sockaddr_in *a)
{
	int s;
	int r=-1;
	socklen_t len;
	struct tcp_info info;
	char addr[INET_ADDRSTRLEN];

	len=sizeof(info);
	memset(&info,0,sizeof(info));
	inet_ntop(AF_INET,&a->sin_addr,addr,sizeof(addr));
	if((s=socket(AF_INET,SOCK_STREAM,0))==-1)perror("socket");
	else if(connect(s,(struct sockaddr *)a,sizeof(struct sockaddr_in)))
		printf("host %s port %d: connect failed\n",addr,
			ntohs(a->sin_port));
	else if(getsockopt(s,IPPROTO_TCP,TCP_INFO,&info,&len))
	{
		perror("getsockopt");
		close(s);
	}
	else
	{
		printf("host %s port %d: pmtu=%d\n",addr,ntohs(a->sin_port),
			info.tcpi_pmtu);
		close(s);
		r=0;
	}
	return r;
}

static int v6test(struct sockaddr_in6 *a)
{
	int s;
	int r=-1;
	socklen_t len;
	struct tcp_info info;
	char addr[INET6_ADDRSTRLEN];

	len=sizeof(info);
	memset(&info,0,sizeof(info));
	inet_ntop(AF_INET6,&a->sin6_addr,addr,sizeof(addr));
	if((s=socket(AF_INET6,SOCK_STREAM,0))==-1)perror("socket");
	else if(connect(s,(struct sockaddr *)a,sizeof(struct sockaddr_in6)))
		printf("host %s port %d: connect failed\n",addr,
			ntohs(a->sin6_port));
	else if(getsockopt(s,IPPROTO_TCP,TCP_INFO,&info,&len))
	{
		perror("getsockopt");
		close(s);
	}
	else
	{
		printf("host %s port %d: pmtu=%d\n",addr,ntohs(a->sin6_port),
			info.tcpi_pmtu);
		close(s);
		r=0;
	}
	return r;
}

static __attribute__((noreturn)) void usage(void)
{
	fprintf(stderr,"Usage: mtucheck <host> <port>\n");
	exit(1);
}

int main(int argc,char *argv[])
{
	int r=0;
	int n=0;
	unsigned long number;
	char *mem;
	struct addrinfo *res;
	struct addrinfo *rp;
	union
	{
		struct sockaddr_in a4;
		struct sockaddr_in6 a6;
	} a;
	struct addrinfo hints;
	char num[6];

	if(argc!=3)usage();
	if(!argv[1][0]||strlen(argv[1])>255)usage();
	if((number=strtoul(argv[2],&mem,10))>65535||!number||*mem)usage();

	if(inet_pton(AF_INET,argv[1],&a.a4.sin_addr.s_addr)==1)
	{
		a.a4.sin_family=AF_INET;
		a.a4.sin_port=htons(number);
		if(v4test(&a.a4))r=1;
		return r;
	}
	else if(inet_pton(AF_INET6,argv[1],&a.a6.sin6_addr.s6_addr)==1)
	{
		a.a6.sin6_family=AF_INET6;
		a.a6.sin6_port=htons(number);
		if(v6test(&a.a6))r=1;
		return r;
	}
	else
	{
		sprintf(num,"%lu",number);
		memset(&hints,0,sizeof(hints));
		hints.ai_family=AF_UNSPEC;
		hints.ai_flags=AI_NUMERICSERV;
		if(!getaddrinfo(argv[1],num,&hints,&res))
		{
			for(rp=res;rp;rp=rp->ai_next)switch(rp->ai_family)
			{
			case AF_INET6:
				n++;
				if(v6test((struct sockaddr_in6 *)rp->ai_addr))
					r=1;
				break;

			case AF_INET:
				n++;
				if(v4test((struct sockaddr_in *)rp->ai_addr))
					r=1;
				break;
			}
			freeaddrinfo(res);
			if(n)return r;
		}
	}

	fprintf(stderr,"Can't find requested address\n");
	return 1;
}
