/*
 * This file is part of the vpnd2 project
 *
 * (C) 2017 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2 or, at
 * your choice, any later version of this license.
 */

#ifndef NO_SCTP
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/sctp.h>
#include <arpa/inet.h>
#include <poll.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#endif
#include "misc.h"
#include "netlink.h"
#include "sctp_oo.h"

COLD int sctpoolisten(void *addr,int port,int keepalive,int sendbuf,int maxseg)
{
#ifndef NO_SCTP
	int s;
	int x;
	int v1;
	int v2;
	int family;
	socklen_t l;
	struct sctp_rtoinfo rto;
	struct sctp_paddrparams ap;
	union
	{
		struct sockaddr_in a4;
		struct sockaddr_in6 a6;
	} a;

	if(port<1||port>65535)goto err1;
	memset(&a,0,sizeof(a));
	if(addr)
	{
		if(inet_pton(AF_INET,addr,&a.a4.sin_addr)==1)
		{
			family=AF_INET;
			a.a4.sin_family=AF_INET;
			a.a4.sin_port=htons(port);
		}
		else if(inet_pton(AF_INET6,addr,&a.a6.sin6_addr)==1)
		{
			family=AF_INET6;
			a.a6.sin6_family=AF_INET6;
			a.a6.sin6_port=htons(port);
		}
		else goto err1;
	}
	else
	{
		family=AF_INET6;
		a.a6.sin6_family=AF_INET6;
		a.a6.sin6_addr=in6addr_any;
		a.a6.sin6_port=htons(port);
	}
	if((s=socket(family,SOCK_STREAM|SOCK_NONBLOCK|SOCK_CLOEXEC,
		IPPROTO_SCTP))==-1)goto err1;
	x=1;
	if(setsockopt(s,SOL_SOCKET,SO_REUSEADDR,&x,sizeof(x)))goto err2;
	x=0;
	if(setsockopt(s,SOL_SOCKET,SO_OOBINLINE,&x,sizeof(x)))goto err2;
	if(family==AF_INET6)
	{
		x=(addr?1:0);
		if(setsockopt(s,IPPROTO_IPV6,IPV6_V6ONLY,&x,sizeof(x)))
			goto err2;
	}
	x=1;
	if(setsockopt(s,SOL_SCTP,SCTP_NODELAY,&x,sizeof(x)))goto err2;
	if(sendbuf)if(setsockopt(s,SOL_SOCKET,SO_SNDBUF,&sendbuf,
		sizeof(sendbuf)))goto err2;
	if(maxseg)if(setsockopt(s,IPPROTO_SCTP,SCTP_MAXSEG,&maxseg,
		sizeof(maxseg)))goto err2;
	l=sizeof(rto);
	memset(&rto,0,sizeof(rto));
	if(getsockopt(s,SOL_SCTP,SCTP_RTOINFO,&rto,&l))goto err2;
	rto.srto_initial=500;
	rto.srto_min=50;
	rto.srto_max=5000;
	if(setsockopt(s,SOL_SCTP,SCTP_RTOINFO,&rto,sizeof(rto)))goto err2;
	if(keepalive)
	{
		v1=SPP_HB_ENABLE;
		v2=SPP_HB_DISABLE;
	}
	else
	{
		v1=SPP_HB_DISABLE;
		v2=SPP_HB_ENABLE;
	}
	l=sizeof(ap);
	memset(&ap,0,sizeof(ap));
	if(getsockopt(s,SOL_SCTP,SCTP_PEER_ADDR_PARAMS,&ap,&l))goto err2;
	ap.spp_hbinterval=keepalive*1000;
	ap.spp_pathmaxrxt=3;
	ap.spp_sackdelay=200;
	ap.spp_flags|=v1|SPP_SACKDELAY_ENABLE;
	ap.spp_flags&=~(v2|SPP_HB_DEMAND|SPP_SACKDELAY_DISABLE|
		SPP_HB_TIME_IS_ZERO);
	if(setsockopt(s,SOL_SCTP,SCTP_PEER_ADDR_PARAMS,&ap,sizeof(ap)))
		goto err2;
	if(bind(s,(struct sockaddr *)(&a),family==AF_INET?
		sizeof(struct sockaddr_in):sizeof(struct sockaddr_in6)))
			goto err2;
	if(listen(s,SOMAXCONN))goto err2;
	return s;

err2:	close(s);
err1:	return -1;
#else
	return -1;
#endif
}

int sctpooaccept(int fd,int keepalive)
{
#ifndef NO_SCTP
	int s;
	int x;
	int v1;
	int v2;
	int family;
	socklen_t ll;
	void *ptr;
	struct sctp_getaddrs *ga;
	struct linger l;
	struct sctp_paddrparams ap;
	union
	{
		struct sockaddr_in a4;
		struct sockaddr_in6 a6;
	} rmt;
	struct sctp_prim pa;
	unsigned char bfr[8192];

	ll=sizeof(rmt);
	if((s=accept4(fd,(struct sockaddr *)&rmt,&ll,
		SOCK_NONBLOCK|SOCK_CLOEXEC))==-1)goto err1;
	if(ll==sizeof(struct sockaddr_in6)&&rmt.a6.sin6_family==AF_INET6)
		family=AF_INET6;
	else if(ll==sizeof(struct sockaddr_in)&&rmt.a4.sin_family==AF_INET)
		family=AF_INET;
	else goto err2;
	x=0;
	if(setsockopt(s,SOL_SOCKET,SO_OOBINLINE,&x,sizeof(x)))goto err2;
	x=1;
	if(setsockopt(s,SOL_SCTP,SCTP_NODELAY,&x,sizeof(x)))goto err2;
	if(keepalive)
	{
		v1=SPP_HB_ENABLE;
		v2=SPP_HB_DISABLE;
	}
	else
	{
		v1=SPP_HB_DISABLE;
		v2=SPP_HB_ENABLE;
	}
	ll=sizeof(ap);
	memset(&ap,0,sizeof(ap));
	if(getsockopt(s,SOL_SCTP,SCTP_PEER_ADDR_PARAMS,&ap,&ll))goto err2;
	ap.spp_hbinterval=keepalive*1000;
	ap.spp_pathmaxrxt=3;
	ap.spp_sackdelay=200;
	ap.spp_flags|=v1|SPP_SACKDELAY_ENABLE;
	ap.spp_flags&=~(v2|SPP_HB_DEMAND|SPP_SACKDELAY_DISABLE|
		SPP_HB_TIME_IS_ZERO);
	if(setsockopt(s,SOL_SCTP,SCTP_PEER_ADDR_PARAMS,&ap,sizeof(ap)))
		goto err2;
	l.l_onoff=0;
	l.l_linger=5;
	if(setsockopt(s,SOL_SOCKET,SO_LINGER,&l,sizeof(l)))goto err2;

	pa.ssp_assoc_id=0;
	memcpy(&pa.ssp_addr,&rmt,family==AF_INET6?
		sizeof(struct sockaddr_in6):sizeof(struct sockaddr_in));
	if(setsockopt(s,SOL_SCTP,SCTP_PRIMARY_ADDR,&pa,sizeof(pa)))goto err2;
	ll=sizeof(bfr);
	ga=(void *)bfr;
	if(!getsockopt(s,SOL_SCTP,SCTP_GET_PEER_ADDRS,bfr,&ll))
		for(x=0,ptr=bfr+sizeof(struct sctp_getaddrs);x<ga->addr_num;x++)
			switch(((struct sockaddr *)ptr)->sa_family)
	{
	case AF_INET6:
		if(family!=AF_INET6||
			!memcmp(&((struct sockaddr_in6 *)ptr)->sin6_addr,
				&((struct sockaddr_in6 *)&rmt)->sin6_addr,16))
					goto n6;
		memset(&ap,0,sizeof(ap));
		memcpy(&ap.spp_address,ptr,sizeof(struct sockaddr_in6));
		ll=sizeof(ap);
		if(getsockopt(s,SOL_SCTP,SCTP_PEER_ADDR_PARAMS,&ap,&ll))goto n6;
		ap.spp_flags|=SPP_HB_DISABLE;
		ap.spp_flags&=~SPP_HB_ENABLE;
		setsockopt(s,SOL_SCTP,SCTP_PEER_ADDR_PARAMS,&ap,sizeof(ap));
n6:		ptr+=sizeof(struct sockaddr_in6);
		break;
	case AF_INET:
		if(family!=AF_INET||
			!memcmp(&((struct sockaddr_in *)ptr)->sin_addr,
				&((struct sockaddr_in *)&rmt)->sin_addr,4))
					goto n4;
		memset(&ap,0,sizeof(ap));
		memcpy(&ap.spp_address,ptr,sizeof(struct sockaddr_in));
		ll=sizeof(ap);
		if(getsockopt(s,SOL_SCTP,SCTP_PEER_ADDR_PARAMS,&ap,&ll))goto n4;
		ap.spp_flags|=SPP_HB_DISABLE;
		ap.spp_flags&=~SPP_HB_ENABLE;
		setsockopt(s,SOL_SCTP,SCTP_PEER_ADDR_PARAMS,&ap,sizeof(ap));
n4:		ptr+=sizeof(struct sockaddr_in);
		break;
	}

	return s;

err2:	close(s);
err1:	return -1;
#else
	return -1;
#endif
}

int sctpooconnect(void *dst,int timeout,int keepalive,int sendbuf,int maxseg,
	int srcaddr,int efd)
{
#ifndef NO_SCTP
	int family;
	int s;
	int x;
	int v1;
	int v2;
	unsigned int y;
	socklen_t ll;
	struct sockaddr *sa;
	void *ptr;
	struct sctp_getaddrs *ga;
	struct linger l;
	struct pollfd p[2];
	struct sctp_rtoinfo rto;
	struct sctp_paddrparams ap;
	struct sctp_prim pa;
	unsigned char bfr[8192];

	family=((struct sockaddr_in *)dst)->sin_family;
	if((s=socket(family,SOCK_STREAM|SOCK_NONBLOCK|SOCK_CLOEXEC,
		IPPROTO_SCTP))==-1)goto err1;
	x=0;
	if(setsockopt(s,SOL_SOCKET,SO_OOBINLINE,&x,sizeof(x)))goto err2;
	x=1;
	if(setsockopt(s,SOL_SCTP,SCTP_NODELAY,&x,sizeof(x)))goto err2;
	ll=sizeof(rto);
	memset(&rto,0,sizeof(rto));
	if(getsockopt(s,SOL_SCTP,SCTP_RTOINFO,&rto,&ll))goto err2;
	rto.srto_initial=500;
	rto.srto_min=50;
	rto.srto_max=5000;
	if(setsockopt(s,SOL_SCTP,SCTP_RTOINFO,&rto,sizeof(rto)))goto err2;
	if(keepalive)
	{
		v1=SPP_HB_ENABLE;
		v2=SPP_HB_DISABLE;
	}
	else
	{
		v1=SPP_HB_DISABLE;
		v2=SPP_HB_ENABLE;
	}
	ll=sizeof(ap);
	memset(&ap,0,sizeof(ap));
	if(getsockopt(s,SOL_SCTP,SCTP_PEER_ADDR_PARAMS,&ap,&ll))goto err2;
	ap.spp_hbinterval=keepalive*1000;
	ap.spp_pathmaxrxt=3;
	ap.spp_sackdelay=200;
	ap.spp_flags|=v1|SPP_SACKDELAY_ENABLE;
	ap.spp_flags&=~(v2|SPP_HB_DEMAND|SPP_SACKDELAY_DISABLE|
		SPP_HB_TIME_IS_ZERO);
	if(setsockopt(s,SOL_SCTP,SCTP_PEER_ADDR_PARAMS,&ap,sizeof(ap)))
		goto err2;
	if(sendbuf)if(setsockopt(s,SOL_SOCKET,SO_SNDBUF,&sendbuf,
		sizeof(sendbuf)))goto err2;
	if(maxseg)if(setsockopt(s,IPPROTO_SCTP,SCTP_MAXSEG,&maxseg,
		sizeof(maxseg)))goto err2;
	l.l_onoff=0;
	l.l_linger=5;
	if(setsockopt(s,SOL_SOCKET,SO_LINGER,&l,sizeof(l)))goto err2;

	if(srcaddr)
	{
		if(!(sa=getsrcaddr(dst,family==AF_INET6?
		    sizeof(struct sockaddr_in6):sizeof(struct sockaddr_in))))
			goto err2;
		if(bind(s,sa,family==AF_INET6?
			sizeof(struct sockaddr_in6):sizeof(struct sockaddr_in)))
		{
			free(sa);
			goto err2;
		}
		free(sa);
	}

	if(connect(s,(struct sockaddr *)dst,family==AF_INET6?
		sizeof(struct sockaddr_in6):sizeof(struct sockaddr_in)))
	{
		if(errno!=EINPROGRESS)goto err2;
		p[0].fd=efd;
		p[0].events=POLLIN;
		p[1].fd=s;
		p[1].events=POLLOUT;
		if(poll(p,2,timeout*1000)<1)goto err2;
		if(p[0].revents&POLLIN)goto err2;
		if(!(p[1].revents&&POLLOUT))goto err2;
		y=sizeof(x);
		if(getsockopt(s,SOL_SOCKET,SO_ERROR,&x,&y))goto err2;
		if(x)goto err2;
	}

	pa.ssp_assoc_id=0;
	memcpy(&pa.ssp_addr,dst,family==AF_INET6?
		sizeof(struct sockaddr_in6):sizeof(struct sockaddr_in));
	if(setsockopt(s,SOL_SCTP,SCTP_PRIMARY_ADDR,&pa,sizeof(pa)))goto err2;
	ll=sizeof(bfr);
	ga=(void *)bfr;
	if(!getsockopt(s,SOL_SCTP,SCTP_GET_PEER_ADDRS,bfr,&ll))
		for(x=0,ptr=bfr+sizeof(struct sctp_getaddrs);x<ga->addr_num;x++)
			switch(((struct sockaddr *)ptr)->sa_family)
	{
	case AF_INET6:
		if(family!=AF_INET6||
			!memcmp(&((struct sockaddr_in6 *)ptr)->sin6_addr,
				&((struct sockaddr_in6 *)dst)->sin6_addr,16))
					goto n6;
		memset(&ap,0,sizeof(ap));
		memcpy(&ap.spp_address,ptr,sizeof(struct sockaddr_in6));
		ll=sizeof(ap);
		if(getsockopt(s,SOL_SCTP,SCTP_PEER_ADDR_PARAMS,&ap,&ll))goto n6;
		ap.spp_flags|=SPP_HB_DISABLE;
		ap.spp_flags&=~SPP_HB_ENABLE;
		setsockopt(s,SOL_SCTP,SCTP_PEER_ADDR_PARAMS,&ap,sizeof(ap));
n6:		ptr+=sizeof(struct sockaddr_in6);
		break;
	case AF_INET:
		if(family!=AF_INET||
			!memcmp(&((struct sockaddr_in *)ptr)->sin_addr,
				&((struct sockaddr_in *)dst)->sin_addr,4))
					goto n4;
		memset(&ap,0,sizeof(ap));
		memcpy(&ap.spp_address,ptr,sizeof(struct sockaddr_in));
		ll=sizeof(ap);
		if(getsockopt(s,SOL_SCTP,SCTP_PEER_ADDR_PARAMS,&ap,&ll))goto n4;
		ap.spp_flags|=SPP_HB_DISABLE;
		ap.spp_flags&=~SPP_HB_ENABLE;
		setsockopt(s,SOL_SCTP,SCTP_PEER_ADDR_PARAMS,&ap,sizeof(ap));
n4:		ptr+=sizeof(struct sockaddr_in);
		break;
	}

	return s;

err2:	close(s);
err1:	return -1;
#else
	return -1;
#endif
}

void sctpooclose(int fd)
{
#ifndef NO_SCTP
	close(fd);
#endif
}

HOT int sctpoosend(int fd,void *data,int len)
{
#ifndef NO_SCTP
	int l;

	if(UNLIKELY((l=write(fd,data,len))==-1))switch(errno)
	{
	case EINTR:
	case EAGAIN:
		return 0;
	default:return -1;
	}
	return l;
#else
	return -1;
#endif
}

HOT int sctpoorecv(int fd,void *data,int max)
{
#ifndef NO_SCTP
	int l;

	if(UNLIKELY((l=read(fd,data,max))==-1))switch(errno)
	{
	case EINTR:
	case EAGAIN:
		return 0;
	default:return -1;
	}
	return l;
#else
	return -1;
#endif
}
