/*
 * This file is part of the vpnd2 project
 *
 * (C) 2017 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2 or, at
 * your choice, any later version of this license.
 */

#include <sys/ioctl.h>
#include <sys/un.h>
#include <sys/socket.h>
#include <poll.h>
#include <stddef.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>

#define BUFSIZE	65536

static int doconnect(char *fn)
{
	int s;
	int x;
	struct sockaddr_un a;

	if(strlen(fn)>=sizeof(a.sun_path))goto err1;
	memset(&a,0,sizeof(a));
	strcpy(a.sun_path,fn);
	a.sun_family=AF_UNIX;
	if((s=socket(AF_UNIX,SOCK_STREAM,0))==-1)goto err1;
	if(connect(s,(struct sockaddr *)&a,sizeof(a)))goto err2;
	x=1;
	if(ioctl(s,FIONBIO,&x))goto err2;
	return s;

err2:	close(s);
err1:	return -1;
}

static void streamforward(int lcl,int rmt)
{
	int len;
	int lrfill;
	int rlfill;
	int lev;
	int rev;
	struct pollfd p[2];
	char lclrmt[BUFSIZE];
	char rmtlcl[BUFSIZE];

	p[0].fd=lcl;
	p[1].fd=rmt;

	lev=POLLIN;
	rev=POLLIN;
	lrfill=0;
	rlfill=0;

	while(lev||rev)
	{
		p[0].events=lev;
		p[1].events=rev;

		while(poll(p,2,-1)<1);

		if(p[0].revents&(POLLERR|POLLHUP|POLLNVAL))break;

		if(p[0].revents&POLLIN)
		{
			if((len=read(lcl,lclrmt+lrfill,BUFSIZE-lrfill))<=0)
				break;
			else lrfill+=len;
		}

		if(p[1].revents&(POLLERR|POLLHUP|POLLNVAL))break;

		if(p[1].revents&POLLIN)
		{
			if((len=read(rmt,rmtlcl+rlfill,BUFSIZE-rlfill))<=0)
				break;
			else rlfill+=len;
		}

		if(lrfill&&(p[1].revents&POLLOUT))
		{
			if((len=write(rmt,lclrmt,lrfill))>=0)
			{
				lrfill-=len;
				if(len)memmove(lclrmt,lclrmt+len,lrfill);
			}
			else break;
		}

		if(rlfill&&(p[0].revents&POLLOUT))
		{
			if((len=write(lcl,rmtlcl,rlfill))>=0)
			{
				rlfill-=len;
				if(len)memmove(rmtlcl,rmtlcl+len,rlfill);
			}
			else break;
		}

		if(lrfill==BUFSIZE)lev=0;
		else lev=POLLIN;
		if(rlfill)lev|=POLLOUT;

		if(rlfill==BUFSIZE)rev=0;
		else rev=POLLIN;
		if(lrfill)rev|=POLLOUT;
	}
}

int main(int argc,char *argv[])
{
	int c;

	close(1);

	if(fcntl(0,F_SETFL,O_RDWR|O_NONBLOCK))goto err1;

	if(argc!=2)goto err1;

	if((c=doconnect(argv[1]))==-1)goto err1;
	if(write(0,"OK\n",3)!=3)goto err2;

	streamforward(0,c);

	close(c);
	close(0);
	return 0;

err2:	close(c);
err1:
	if(write(0,"NO\n",3)!=3);
	return 1;
}
