/*
 * This file is part of the vpnd2 project
 *
 * (C) 2020 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2 or, at
 * your choice, any later version of this license.
 */

#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <errno.h>
#include <poll.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include "httpconn.h"
#include "network.h"

int httpconnect(struct sockaddr *server,struct sockaddr *dest,char *host,
	int port,int contimeout,int efd,int idle,int fast)
{
	int s;
	int len;
	int e;
	socklen_t l;
	struct sockaddr_in *a4=(void *)dest;
	struct sockaddr_in6 *a6=(void *)dest;
	struct pollfd p[2];
	char addr[64];
	char req[1024];

	if(!host)switch(dest->sa_family)
	{
	case AF_INET:
		if(!inet_ntop(AF_INET,&a4->sin_addr,addr,sizeof(addr)))
			return -1;
		host=addr;
		port=ntohs(a4->sin_port);

	case AF_INET6:
		if(!inet_ntop(AF_INET,&a6->sin6_addr,addr+1,sizeof(addr)-2))
			return -1;
		addr[0]='[';
		strcat(addr,"]");
		host=addr;
		port=ntohs(a6->sin6_port);
		break;

	default:return -1;
	}
	else if(inet_pton(AF_INET6,host,addr)==1)
	{
		addr[0]='[';
		strcpy(addr+1,host);
		strcat(addr,"]");
		host=addr;
	}

	if(port<1||port>65535)return -1;

	if((len=snprintf(req,sizeof(req),"CONNECT %s:%u HTTP/1.1\r\n"
		"HOST: %s:%u\r\n\r\n",host,port,host,port))==sizeof(req)-1)
			return -1;

	if((s=tcpconnect(NULL,server,5,efd,
		(idle?KEEPALIVE:0)|(fast?FAST_CONNECT:0),NULL,idle))<0)return s;

	p[0].fd=s;
	p[0].events=POLLOUT;
	p[1].fd=efd;
	p[1].events=(efd==-1?0:POLLIN);
	p[1].revents=0;

	for(e=0;e<len;)
	{
		switch(poll(p,efd==-1?1:2,500))
		{
		case 0: close(s);
			return -1;
		case -1:continue;
		}
		if(p[1].revents)
		{
			close(s);
			return -2;
		}
		if(!(p[0].revents&POLLOUT))
		{
			close(s);
			return -1;
		}
		if((l=write(s,req+e,len-e))<=0)
		{
			close(s);
			return -1;
		}
		e+=l;
	}

	switch(poll(p,efd==-1?1:2,contimeout*1000))
	{
	case -1:
	case 0: close(s);
		return -1;
	}

	for(p[0].events=POLLIN,e=0,l=0,len=0;;)
	{
		if(e==sizeof(req))
		{
			close(s);
			return 1;
		}
		switch(poll(p,efd==-1?1:2,500))
		{
		case 0: close(s);
			return -1;
		case -1:continue;
		}
		if(p[1].revents)
		{
			close(s);
			return -2;
		}
		if(!(p[0].revents&POLLIN))
		{
			close(s);
			return -1;
		}
		if(read(s,req+e,1)!=1)
		{
			close(s);
			return 1;
		}
		switch(req[e])
		{
		case '\n':
			if(!l++)
			{
				if(strncmp(req,"HTTP/1.0 200 ",13)&&
					strncmp(req,"HTTP/1.1 200 ",13))
				{
					close(s);
					return -1;
				}
			}
			if(++len==2)return s;
			e++;
			break;

		default:e++;
			len=0;
		case '\r':
			break;
		}
	}
}
