/*
 * This file is part of the vpnd2 project
 *
 * (C) 2020 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2 or, at
 * your choice, any later version of this license.
 */

#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <poll.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include "s4client.h"
#include "network.h"

int s4connect(struct sockaddr *server,char *user,struct sockaddr_in *dest,
	char *host,int contimeout,int efd,int idle,int fast)
{
	int s;
	int port;
	int ulen;
	int hlen;
	int rlen;
	int e;
	struct pollfd p[2];
	unsigned char bfr[520];

	if((ulen=strlen(user))>255)
	{
		fprintf(stderr,"oversize user name\n");
		return -1;
	}

	if(host)if((hlen=strlen(host))>255)
	{
		fprintf(stderr,"oversize host name\n");
		return -1;
	}

	port=ntohs(dest->sin_port);

	bfr[0]=0x04;
	bfr[1]=0x01;
	bfr[2]=(unsigned char)(port>>8);
	bfr[3]=port&0xff;
	if(host)
	{
		bfr[4]=0x00;
		bfr[5]=0x00;
		bfr[6]=0x00;
		bfr[7]=0x01;
	}
	else memcpy(bfr+4,&dest->sin_addr.s_addr,4);
	strcpy((char *)bfr+8,user);
	if(host)strcpy((char *)bfr+ulen+9,host);
	rlen=9+ulen+(host?hlen+1:0);

	if((s=tcpconnect(NULL,server,5,efd,
		(idle?KEEPALIVE:0)|(fast?FAST_CONNECT:0),NULL,idle))<0)
			return s;

	p[0].fd=s;
	p[0].events=POLLIN;
	p[1].fd=efd;
	p[1].events=(efd==-1?0:POLLIN);
	p[1].revents=0;

	if(write(s,bfr,rlen)!=rlen)
	{
		perror("write");
		close(s);
		return -1;
	}

	if(poll(p,efd==-1?1:2,contimeout*1000)<1||p[1].revents||!p[0].revents)
	{
		close(s);
		if(p[1].revents)
		{
			fprintf(stderr,"abort\n");
			return -2;
		}
		fprintf(stderr,"timeout\n");
		return -1;
	}

	for(rlen=0;rlen<8;rlen+=e)if((e=read(s,bfr+rlen,8-rlen))<=0)
	{
		if(e==-1&&errno==EINTR)
		{
			e=0;
			continue;
		}
		perror("read");
		close(s);
		return -1;
	}

	if(bfr[0]!=0x00||(bfr[1]!=0x5a&&bfr[1]!=0x5b))
	{
		fprintf(stderr,"format error\n");
		close(s);
		return -1;
	}

	if(bfr[1]==0x5b)
	{
		fprintf(stderr,"request denied by server\n");
		close(s);
		return -1;
	}

	memcpy(&dest->sin_port,bfr+2,2);
	memcpy(&dest->sin_addr.s_addr,bfr+4,4);

	return s;
}
