/*
 * This file is part of the vpnd2 project
 *
 * (C) 2017 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2 or, at
 * your choice, any later version of this license.
 */

#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>
#include <errno.h>
#include "misc.h"
#include "unix.h"

COLD int unix_listen(void *fn,int packet)
{
	int s;
	mode_t m;
	struct stat stb;
	struct sockaddr_un a;

	if(strlen(fn)>=sizeof(a.sun_path))goto err1;
	if(!stat(fn,&stb)&&S_ISSOCK(stb.st_mode))if(unlink(fn))goto err1;
	if((s=socket(AF_UNIX,
		(packet?SOCK_SEQPACKET:SOCK_STREAM)|SOCK_NONBLOCK|SOCK_CLOEXEC,
		0))==-1)goto err1;
	memset(&a,0,sizeof(a));
	strcpy(a.sun_path,fn);
	a.sun_family=AF_UNIX;
	m=umask(0);
	if(bind(s,(struct sockaddr *)&a,sizeof(a)))goto err2;
	umask(m);
	if(listen(s,SOMAXCONN))goto err3;
	return s;

err3:	unlink(fn);
err2:	close(s);
err1:	return -1;
}

COLD void unix_clean(void *fn)
{
	struct stat stb;

	if(!stat(fn,&stb)&&S_ISSOCK(stb.st_mode))unlink(fn);
}

int unix_accept(int fd)
{
	return accept4(fd,NULL,NULL,SOCK_NONBLOCK|SOCK_CLOEXEC);
}

void unix_close(int fd)
{
	close(fd);
}

HOT int unixsend(int fd,void *data,int len)
{
	int l;

	if(UNLIKELY((l=write(fd,data,len))==-1))switch(errno)
	{
	case EINTR:
	case EAGAIN:
		return 0;
	default:return -1;
	}
	return l;
}

HOT int unixrecv(int fd,void *data,int max)
{
	int l;

	if(UNLIKELY((l=read(fd,data,max))==-1))switch(errno)
	{
	case EINTR:
	case EAGAIN:
		return 0;
	default:return -1;
	}
	return l;
}

HOT int unixpktsend(int fd,void *data,int len)
{
	int l;

	if(UNLIKELY((l=send(fd,data,len,MSG_EOR))==-1))switch(errno)
	{
	case EINTR:
	case EAGAIN:
		return 0;
	default:return -1;
	}
	if(UNLIKELY(l!=len))return -1;
	return l;
}

HOT int unixpktrecv(int fd,void *data,int max)
{
	int l;

	if(UNLIKELY((l=recv(fd,data,max,MSG_NOSIGNAL|MSG_WAITALL))==-1))
		switch(errno)
	{
	case EINTR:
	case EAGAIN:
		return 0;
	default:return -1;
	}
	return l;
}
