/*
 * This file is part of the vpnd2 project
 *
 * (C) 2017 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2 or, at
 * your choice, any later version of this license.
 */

#include "crypto.h"
#include "keyio.h"
#include <string.h>
#include <stdio.h>

int generate(char *fn,char *which,void *glob,void *passwd,int plen)
{
	int curve;
	void *h;

	if(!strcmp(which,"secp521r1"))curve=CRYPTO_SECP521R1;
	else if(!strcmp(which,"secp384r1"))curve=CRYPTO_SECP384R1;
	else if(!strcmp(which,"secp256r1"))curve=CRYPTO_SECP256R1;
	else if(!strcmp(which,"brainpoolp512r1"))curve=CRYPTO_BRAINPOOLP512R1;
	else if(!strcmp(which,"brainpoolp384r1"))curve=CRYPTO_BRAINPOOLP384R1;
	else if(!strcmp(which,"brainpoolp256r1"))curve=CRYPTO_BRAINPOOLP256R1;
	else goto err;

	if(!(h=crypto_thread_init(glob)))goto err;
	if(genkey(h,fn,curve,passwd,plen))
	{
		crypto_thread_exit(h);
		goto err;
	}
	crypto_thread_exit(h);
	return 0;

err:	fprintf(stderr,"key generate failed\n");
	return -1;
}
